#include "FFGLShader.h"
#include <stdio.h>
#include <vector>
#include <assert.h>
#include "FFGLUtilities.h"

#define LOGSHADERERRORS

namespace ffglex
{
/**
 * The default constructor just initializes this object to represent no shader.
 */
FFGLShader::FFGLShader() :
	vertexShaderID( 0 ),
	geometryShaderID( 0 ),
	fragmentShaderID( 0 ),
	programID( 0 )
{
}
/**
 * Destructing this object just checks and lets you know if you've properly released the GL resources.
 */
FFGLShader::~FFGLShader()
{
	//If any of these assertions hit you forgot to free this shader's gl resources.
	assert( vertexShaderID == 0 );
	assert( geometryShaderID == 0 );
	assert( fragmentShaderID == 0 );
	assert( programID == 0 );
}

/**
 * Add a varying that will be captured using a transform feedback.
 * If you want to use transform feedback shaders you have to specify which varyings you want to capture.
 * You can do this by calling this function for each varying you want to capture prior to compiling the shader. You should
 * call this function in order of the varyings to capture.
 *
 * @param varyingName: The name of the varying that you want to capture.
 */
void FFGLShader::AddTransformFeedbackVarying( const std::string& varyingName )
{
	transformFeedbackVaryings.push_back( varyingName );
}
/**
 * Compiles and links two shaders into a shader program.
 *
 * @param vertexShader: The glsl shader string that represents the shader that should be used in the vertex processing stage.
 * You should make sure that this glsl string defines the glsl version it's targetting (eg #version 330) and respect the OpenGL specs
 * for that shader version.
 * @param fragmentShader: The glsl shader string that represents the shader that should be used in the fragment processing stage.
 * You should make sure that this glsl string defines the glsl version it's targetting (eg #version 330) and respect the OpenGL specs
 * for that shader version.
 * @return: Whether or not compiling/linking succeeded and thus if this object can be used to bind a shader program for rendering.
 * If compilation succeeded you need to make sure that you're freeing the GL resources when you no longer intend to use this shader.
 */
bool FFGLShader::Compile( const char* vertexShader, const char* fragmentShader )
{
	return Compile( vertexShader, nullptr, fragmentShader );
}
/**
 * Compiles and links a vertex/geometry/fragment shader into a single shader program.
 *
 * @param vertexShader: The glsl shader string that represents the shader that should be used in the vertex processing stage.
 * You should make sure that this glsl string defines the glsl version it's targetting (eg #version 330) and respect the OpenGL specs
 * for that shader version.
 * @param geometryShader: The glsl shader string that represents the shader that should be used in the geometry processing stage.
 * You should make sure that this glsl string defines the glsl version it's targetting (eg #version 330) and respect the OpenGL specs
 * for that shader version.
 * @param fragmentShader: The glsl shader string that represents the shader that should be used in the fragment processing stage.
 * You should make sure that this glsl string defines the glsl version it's targetting (eg #version 330) and respect the OpenGL specs
 * for that shader version.
 * @return: Whether or not compiling/linking succeeded and thus if this object can be used to bind a shader program for rendering.
 * If compilation succeeded you need to make sure that you're freeing the GL resources when you no longer intend to use this shader.
 */
bool FFGLShader::Compile( const char* vertexShader, const char* geometryShader, const char* fragmentShader )
{
	//We need to have valid inputs to be able to compile a shader, validate them here so that below we can just use them without checking.
	//All shaders require a vertex shader.
	//No shaders require the geometry stage.
	//Transform feedback shaders dont require a fragment shader.
	if( vertexShader == nullptr || strlen( vertexShader ) == 0 )
		return false;
	if( ( fragmentShader == nullptr || strlen( fragmentShader ) == 0 ) && transformFeedbackVaryings.empty() )
		return false;

	// Structure for fragment, geometry and vertex shader information to loop over
	struct shaderInfo
	{
		GLuint& ID;// Note: ID will be updated
		const GLenum type;
		const GLchar* const code;
		const char* const name;
	} shaders[] =
		{
			{ vertexShaderID, GL_VERTEX_SHADER, vertexShader, "Vertex" },
			{ geometryShaderID, GL_GEOMETRY_SHADER, geometryShader, "Geometry" },
			{ fragmentShaderID, GL_FRAGMENT_SHADER, fragmentShader, "Fragment" }
		};

	const int numShaders = sizeof( shaders ) / sizeof( *shaders );

	for( int i = 0; i < numShaders; i++ )
	{
		const shaderInfo& shader = shaders[ i ];

		if( shader.code != nullptr && strlen( shader.code ) != 0 )
		{
			// Create the shader and store the returned ID.
			shader.ID = glCreateShader( shader.type );

			if( !CompileShader( shader.ID, shader.code, shader.name ) )
			{
				FreeGLResources();
				return false;
			}
		}
	}

	if( !LinkProgram() )
	{
		FreeGLResources();
		return false;
	}

	return true;
}
/**
 * Frees any GL resources that this shader is holding.
 *
 * Call this when you no longer intend to use this shader.
 */
void FFGLShader::FreeGLResources()
{
	if( vertexShaderID != 0 )
	{
		glDeleteShader( vertexShaderID );
		vertexShaderID = 0;
	}

	if( fragmentShaderID != 0 )
	{
		glDeleteShader( fragmentShaderID );
		fragmentShaderID = 0;
	}

	if( geometryShaderID != 0 )
	{
		glDeleteShader( geometryShaderID );
		geometryShaderID = 0;
	}

	if( programID != 0 )
	{
		glDeleteProgram( programID );
		programID = 0;
	}
}

/**
 * Whether or not previous complilation succeeded and this shader is ready to be used for rendering.
 */
bool FFGLShader::IsReady() const
{
	//The shaders themselves are optional, it's the program we're using when rendering with those shaders.
	return programID != 0;
}
void FFGLShader::Use() const
{
	glUseProgram( programID );
}
/**
 * Gets the OpenGL ID that represents the linked shader program.
 *
 * If the shader is usable this returns anything other than 0.
 */
GLuint FFGLShader::GetGLID() const
{
	return programID;
}
/**
 * Finds the uniform location that needs to be used to addess a uniform with a certain name.
 *
 * @param name: The name of the uniform for which you want to find the location. This name has to be
 * a uniform that's placed inside any of the shader strings provided to the previous successfull call to Compile.
 * @return: The uniform's location to be passed into a glUniform{1|2|3|4}{f|i} function call. -1 if no uniform with the
 * provided name is available.
 */
GLint FFGLShader::FindUniform( const char* name ) const
{
	return glGetUniformLocation( programID, name );
}

void FFGLShader::Set( const char* name, float value )
{
	glUniform1f( FindUniform( name ), value );
}

void FFGLShader::Set( const char* name, float v1, float v2 )
{
	glUniform2f( FindUniform( name ), v1, v2 );
}

void FFGLShader::Set( const char* name, float v1, float v2, float v3 )
{
	glUniform3f( FindUniform( name ), v1, v2, v3 );
}

void FFGLShader::Set( const char* name, float v1, float v2, float v3, float v4 )
{
	glUniform4f( FindUniform( name ), v1, v2, v3, v4 );
}

void FFGLShader::Set( const char* name, int value )
{
	glUniform1i( FindUniform( name ), value );
}

void FFGLShader::Bind( const char* name, int textureID, const FFGLTextureStruct& texture )
{
	Set( name, textureID );
	glActiveTexture( GL_TEXTURE0 + textureID );
	glBindTexture( GL_TEXTURE_2D, texture.Handle );
}

bool FFGLShader::CompileShader( const GLuint shaderID, const char* shaderCode, const char* shaderName )
{
	// Load Shader Source
	glShaderSource( shaderID, 1, &shaderCode, NULL );

	// Compile The Shader
	glCompileShader( shaderID );

	// Check for log message (can be a warning message even if compilation succeeds)
	{
		GLint logLength;
		glGetShaderiv( shaderID, GL_INFO_LOG_LENGTH, &logLength );

		if( logLength > 0 )
		{
			// Get the log so we can peek at the error string
			GLchar* logStr = new GLchar[ logLength + 1 ];
			glGetShaderInfoLog( shaderID, logLength, NULL, logStr );
#ifdef LOGSHADERERRORS
			printf( "%s shader log: %s\n", shaderName, logStr );
#endif
			delete[] logStr;
		}
	}

	// Check for compilation success
	GLint compileStatus;
	glGetShaderiv( shaderID, GL_COMPILE_STATUS, &compileStatus );

	return compileStatus == GL_TRUE;
}
bool FFGLShader::LinkProgram()
{
	programID = glCreateProgram();

	glAttachShader( programID, vertexShaderID );
	if( geometryShaderID != 0 )
		glAttachShader( programID, geometryShaderID );
	if( fragmentShaderID != 0 )
		glAttachShader( programID, fragmentShaderID );

	if( !transformFeedbackVaryings.empty() )
	{
		std::vector< const char* > feedbackVaryingNames;
		for( size_t index = 0; index < transformFeedbackVaryings.size(); ++index )
			feedbackVaryingNames.push_back( transformFeedbackVaryings[ index ].c_str() );
		glTransformFeedbackVaryings( programID, (GLsizei)feedbackVaryingNames.size(), feedbackVaryingNames.data(), GL_INTERLEAVED_ATTRIBS );
	}

	GLint linkStatus = 0;
	glLinkProgram( programID );
	glGetProgramiv( programID, GL_LINK_STATUS, &linkStatus );

	if( linkStatus != GL_TRUE )
	{
#ifdef LOGSHADERERRORS
		//get the log so we can peek at the error string
		char log[ 1024 ];
		GLsizei returnedLength;
		glGetProgramInfoLog( fragmentShaderID, sizeof( log ) - 1, &returnedLength, log );
		log[ returnedLength ] = 0;

		Log( "Fragment Shader error: ", log );
#endif
	}

	return linkStatus == GL_TRUE;
}

}//End namespace ffglex